import grp
import os
import pwd
import shutil
import stat
import sys
import tempfile
import unittest
from unittest import mock


# We also need to mock up charms.apt and charms.layer so we can run
# unit tests without having to build the charm and pull in layers such
# as layer-status.
sys.modules['charms.apt'] = mock.MagicMock()
sys.modules['charms.layer'] = mock.MagicMock()
from charms import apt  # NOQA: E402
from charms.layer import status  # NOQA: E402

# Add path to where our reactive layer lives and import.
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))
from reactive import jenkins_slave  # NOQA: E402


INITIAL_CONF = 'tests/unit/files/jenkins-slave-default'


class TestSetDefaultConf(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.tmpdir = tempfile.mkdtemp(prefix='charm-unittests-')
        self.addCleanup(shutil.rmtree, self.tmpdir)
        temp_file = tempfile.NamedTemporaryFile(delete=False, dir=self.tmpdir)
        with open(INITIAL_CONF, 'rb') as f:
            conf = f.read().decode('utf-8')
        temp_file.write(conf.encode())
        temp_file.close()
        self.conf_file = temp_file.name
        self.user = pwd.getpwuid(os.getuid()).pw_name
        self.group = grp.getgrgid(os.getgid()).gr_name
        self.charm_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

        # charmhelpers is getting difficult to test against, as it writes
        # to system directories even for things that should be idempotent,
        # like accessing config options.
        patcher = mock.patch('charmhelpers.core.hookenv.charm_dir')
        self.mock_charm_dir = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_charm_dir.return_value = self.charm_dir

        patcher = mock.patch('charmhelpers.core.hookenv.config')
        self.mock_config = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_config.return_value = {'tools': []}

        patcher = mock.patch('charmhelpers.core.hookenv.local_unit')
        self.mock_local_unit = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_local_unit.return_value = 'mock-jenkins-slave/0'

        patcher = mock.patch('charmhelpers.core.hookenv.unit_private_ip')
        self.mock_local_unit = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_local_unit.return_value = '10.1.2.3'

        patcher = mock.patch('charmhelpers.core.hookenv.log')
        self.mock_log = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_log.return_value = ''

        patcher = mock.patch('os.uname')
        self.mock_log = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_log.return_value = (
            'Linux',
            'mock-jenkins-slave',
            '4.15.0-46-generic',
            '#49-Ubuntu SMP Wed Feb 6 09:33:07 UTC 2019',
            'x86_64',
        )

        patcher = mock.patch('os.cpu_count')
        self.mock_log = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_log.return_value = 3

        apt.reset_mock()
        status.active.reset_mock()
        status.blocked.reset_mock()
        status.maintenance.reset_mock()

    @mock.patch('charms.reactive.clear_flag')
    def test_hook_upgrade_charm_flags(self, clear_flag):
        '''Test correct flags set via upgrade-charm hook'''
        jenkins_slave.upgrade_charm()
        self.assertFalse(status.maintenance.assert_called())
        expected = [
            mock.call('jenkins-slave.active'),
            mock.call('jenkins-slave.installed'),
            mock.call('nagios-nrpe.configured'),
        ]
        self.assertFalse(clear_flag.assert_has_calls(expected, any_order=True))

    @mock.patch('charms.reactive.clear_flag')
    def test_hook_config_changed(self, clear_flag):
        '''Test correct flags are set via config-changed charm hook'''
        jenkins_slave.config_changed()
        self.assertFalse(status.maintenance.assert_called())
        expected = [mock.call('jenkins-slave.configured'), mock.call('nagios-nrpe.configured')]
        self.assertFalse(clear_flag.assert_has_calls(expected, any_order=True))

    @mock.patch('charmhelpers.core.hookenv.config')
    @mock.patch('charmhelpers.core.host.adduser')
    @mock.patch('charmhelpers.core.host.mkdir')
    @mock.patch('charmhelpers.core.host.service')
    @mock.patch('charmhelpers.core.host.user_exists')
    @mock.patch('reactive.jenkins_slave.apt_purge')
    @mock.patch('reactive.jenkins_slave.file_to_units')
    @mock.patch('reactive.jenkins_slave.write_default_conf')
    def test_hook_install(
        self, write_default_conf, file_to_units, apt_purge, user_exists, service, mkdir, adduser, config
    ):
        config.return_value = {'tools': 'some-tools-package'}
        apt.install_queued.return_value = True
        user_exists.return_value = False
        jenkins_slave.install()
        self.assertFalse(status.maintenance.assert_called())

        expected = [mock.call(home_dir='/var/lib/jenkins', system_user=True, username='jenkins')]
        self.assertFalse(adduser.assert_has_calls(expected, any_order=True))
        expected = [
            mock.call('/var/lib/jenkins', group='jenkins', owner='jenkins', perms=0o755),
            mock.call('/var/log/jenkins', group='jenkins', owner='jenkins', perms=0o755),
        ]
        self.assertFalse(mkdir.assert_has_calls(expected, any_order=True))
        self.assertFalse(service.assert_has_calls([mock.call('enable', 'jenkins-slave')], any_order=True))
        self.assertFalse(apt_purge.assert_has_calls([mock.call(['jenkins-slave'])], any_order=True))
        expected = [
            mock.call('files/download-slave.sh', '/usr/local/sbin/download-slave.sh'),
            mock.call('files/jenkins-slave-logrotate-config', '/etc/logrotate.d/jenkins-slave'),
            mock.call('files/jenkins-slave-systemd-config', '/lib/systemd/system/jenkins-slave.service'),
        ]
        self.assertFalse(file_to_units.assert_has_calls(expected, any_order=True))
        self.assertFalse(write_default_conf.assert_called())
        expected = [
            mock.call.queue_install(['wget', 'default-jre-headless', 'some-tools-package']),
            mock.call.install_queued(),
        ]
        self.assertEqual(apt.method_calls, expected)

        apt.install_queued.return_value = False
        apt_purge.reset_mock()
        jenkins_slave.install()
        self.assertFalse(apt_purge.assert_not_called())

        apt.install_queued.return_value = True
        user_exists.return_value = True
        adduser.reset_mock()
        jenkins_slave.install()
        self.assertFalse(adduser.assert_not_called())

    @mock.patch('charmhelpers.core.host.lsb_release')
    @mock.patch('charmhelpers.core.host.adduser')
    @mock.patch('charmhelpers.core.host.mkdir')
    @mock.patch('reactive.jenkins_slave.file_to_units')
    @mock.patch('reactive.jenkins_slave.write_default_conf')
    def test_hook_install_trusty(self, write_default_conf, file_to_units, mkdir, adduser, lsb_release):
        apt.install_queued.return_value = True
        lsb_release.return_value = {'DISTRIB_CODENAME': 'trusty'}
        jenkins_slave.install()
        self.assertFalse(status.maintenance.assert_called())

        file_to_units('files/jenkins-slave-upstart-config', '/etc/init/jenkins-slave.conf')
        expected = [mock.call('files/jenkins-slave-upstart-config', '/etc/init/jenkins-slave.conf')]
        self.assertFalse(file_to_units.assert_has_calls(expected, any_order=True))

    @mock.patch('charms.reactive.clear_flag')
    @mock.patch('charms.reactive.set_flag')
    @mock.patch('charmhelpers.core.hookenv.config')
    @mock.patch('charmhelpers.core.unitdata.kv')
    @mock.patch('reactive.jenkins_slave.write_default_conf')
    @mock.patch('reactive.jenkins_slave.file_to_units')
    def test_configure_jenkins_slave_no_url(
        self, file_to_units, write_default_conf, unitdata_kv, config, set_flag, clear_flag
    ):
        config.return_value = {}
        unitdata_kv.return_value = {}
        jenkins_slave.configure_jenkins_slave()
        self.assertFalse(write_default_conf.assert_not_called())
        self.assertFalse(status.blocked.assert_called())
        self.assertFalse(clear_flag.assert_has_calls([mock.call('jenkins-slave.active')], any_order=True))

    @mock.patch('charms.reactive.clear_flag')
    @mock.patch('charms.reactive.set_flag')
    @mock.patch('charmhelpers.core.hookenv.config')
    @mock.patch('charmhelpers.core.unitdata.kv')
    @mock.patch('reactive.jenkins_slave.write_default_conf')
    @mock.patch('reactive.jenkins_slave.file_to_units')
    def test_configure_jenkins_slave_master_url(
        self, file_to_units, write_default_conf, unitdata_kv, config, set_flag, clear_flag
    ):
        config.return_value = {'master_url': 'http://10.1.1.1:8080'}
        unitdata_kv.return_value = {}
        jenkins_slave.configure_jenkins_slave()
        self.assertFalse(write_default_conf.assert_called_once_with('http://10.1.1.1:8080'))
        self.assertFalse(set_flag.assert_has_calls([mock.call('jenkins-slave.configured')], any_order=True))
        expected = [mock.call('jenkins-slave.active'), mock.call('nagios-nrpe.configured')]
        self.assertFalse(clear_flag.assert_has_calls(expected, any_order=True))

    @mock.patch('charms.reactive.clear_flag')
    @mock.patch('charms.reactive.set_flag')
    @mock.patch('charmhelpers.core.hookenv.config')
    @mock.patch('charmhelpers.core.unitdata.kv')
    @mock.patch('reactive.jenkins_slave.write_default_conf')
    @mock.patch('reactive.jenkins_slave.file_to_units')
    def test_configure_jenkins_slave_relation_url(
        self, file_to_units, write_default_conf, unitdata_kv, config, set_flag, clear_flag
    ):
        config.return_value = {}
        unitdata_kv.return_value = {'url': 'http://10.22.22.22:8080'}
        jenkins_slave.configure_jenkins_slave()
        self.assertTrue(write_default_conf.called_once_with('http://10.22.22.22:8080'))
        self.assertFalse(set_flag.assert_has_calls([mock.call('jenkins-slave.configured')], any_order=True))
        expected = [mock.call('jenkins-slave.active'), mock.call('nagios-nrpe.configured')]
        self.assertFalse(clear_flag.assert_has_calls(expected, any_order=True))

    @mock.patch('charmhelpers.core.host.service_running')
    @mock.patch('charmhelpers.core.host.service_restart')
    @mock.patch('charmhelpers.core.host.service_start')
    def test_set_active_running(self, service_start, service_restart, service_running):
        '''Test service restarted when already running'''
        service_running.return_value = True
        jenkins_slave.set_active()
        self.assertFalse(status.maintenance.assert_called())
        self.assertFalse(service_start.assert_not_called())
        self.assertFalse(service_restart.assert_called_once_with('jenkins-slave'))

    @mock.patch('charmhelpers.core.host.service_running')
    @mock.patch('charmhelpers.core.host.service_restart')
    @mock.patch('charmhelpers.core.host.service_start')
    def test_set_active_not_running(self, service_start, service_restart, service_running):
        '''Test service restarted when not running'''
        service_running.return_value = False
        jenkins_slave.set_active()
        self.assertFalse(status.maintenance.assert_called())
        self.assertFalse(service_start.assert_called_once_with('jenkins-slave'))
        self.assertFalse(service_restart.assert_not_called())

    @mock.patch('charms.reactive.set_flag')
    @mock.patch('charmhelpers.core.hookenv.config')
    @mock.patch('charmhelpers.core.unitdata.kv')
    @mock.patch('charmhelpers.core.host.service_running')
    @mock.patch('charmhelpers.core.host.service_start')
    def test_set_active_master_url_or_relation(self, service_start, service_running, unitdata_kv, config, set_flag):
        service_running.return_value = False
        config.return_value = {'master_url': 'http://ext-jenkins-be.internal:8080'}
        jenkins_slave.set_active()
        status.active.assert_called_once_with('ready - using jenkins http://ext-jenkins-be.internal:8080 (master_url)')

        config.return_value = {}
        status.active.reset_mock()
        jenkins_slave.set_active()
        status.active.assert_called_once_with('ready - using jenkins from relation')

    @mock.patch('charms.reactive.clear_flag')
    @mock.patch('charms.reactive.set_flag')
    def test_hook_slave_relation_changed_flags(self, set_flag, clear_flag):
        jenkins_slave.slave_relation_changed()
        expected = [mock.call('jenkins-slave.configured'), mock.call('slave-relation.configured')]
        self.assertFalse(clear_flag.assert_has_calls(expected, any_order=True))
        self.assertFalse(set_flag.assert_has_calls([mock.call('slave-relation.available')], any_order=True))

    @mock.patch('charms.reactive.clear_flag')
    def test_hook_slave_relation_removed_flags(self, clear_flag):
        jenkins_slave.slave_relation_removed()
        self.assertFalse(clear_flag.assert_has_calls([mock.call('slave-relation.available')], any_order=True))

    @mock.patch('charms.reactive.clear_flag')
    @mock.patch('charms.reactive.set_flag')
    @mock.patch('charmhelpers.core.hookenv.config')
    @mock.patch('charmhelpers.core.hookenv.relation_get')
    @mock.patch('charmhelpers.core.hookenv.relation_set')
    @mock.patch('charmhelpers.core.unitdata.kv')
    @mock.patch('reactive.jenkins_slave.write_default_conf')
    def test_hook_slave_relation(
        self, write_default_conf, unitdata_kv, relation_set, relation_get, config, set_flag, clear_flag
    ):
        config.return_value = {'master_url': ''}
        relation_get.return_value = 'http://10.1.1.1:8080'
        jenkins_slave.slave_relation()
        self.assertFalse(relation_get.assert_called_once_with('url'))
        self.assertFalse(clear_flag.assert_called_once_with('jenkins-slave.active'))
        self.assertFalse(set_flag.assert_called_once_with('slave-relation.configured'))
        expected = [
            mock.call(executors=3),
            mock.call(labels='x86_64'),
            mock.call(slavehost='mock-jenkins-slave-0'),
            mock.call(slaveaddress='10.1.2.3'),
        ]
        self.assertFalse(relation_set.assert_has_calls(expected, any_order=True))

        config.return_value = {'master_url': '', 'labels': 'label1 label2'}
        relation_get.return_value = 'http://10.1.1.1:8080'
        relation_set.reset_mock()
        jenkins_slave.slave_relation()
        expected = [
            mock.call(executors=3),
            mock.call(labels='label1 label2'),
            mock.call(slavehost='mock-jenkins-slave-0'),
            mock.call(slaveaddress='10.1.2.3'),
        ]
        self.assertFalse(relation_set.assert_has_calls(expected, any_order=True))

    @mock.patch('charmhelpers.core.hookenv.config')
    @mock.patch('charmhelpers.core.hookenv.relation_get')
    @mock.patch('charms.reactive.set_flag')
    def test_hook_slave_relation_master_url_set(self, set_flag, relation_get, config):
        config.return_value = {'master_url': 'http://10.1.1.1:8080'}
        relation_get.return_value = ''
        jenkins_slave.slave_relation()
        self.assertFalse(relation_get.assert_not_called())
        self.assertFalse(set_flag.assert_called_once_with('slave-relation.configured'))

    @mock.patch('charms.reactive.clear_flag')
    @mock.patch('charms.reactive.set_flag')
    @mock.patch('charmhelpers.core.hookenv.config')
    @mock.patch('charmhelpers.core.hookenv.relation_get')
    def test_hook_slave_relation_relation_url_not_set(self, relation_get, config, set_flag, clear_flag):
        config.return_value = {'master_url': ''}
        relation_get.return_value = ''
        jenkins_slave.slave_relation()
        self.assertFalse(clear_flag.assert_not_called())

    @mock.patch('charms.reactive.set_flag')
    @mock.patch('charmhelpers.contrib.charmsupport.nrpe.get_nagios_hostname')
    @mock.patch('charmhelpers.contrib.charmsupport.nrpe.NRPE')
    def test_configure_nagios(self, nrpe, get_nagios_hostname, set_flag):
        get_nagios_hostname.return_value = 'some-host.local'
        nrpe_instance_mock = nrpe(get_nagios_hostname(), primary=True)
        jenkins_slave.configure_nagios(None)
        status.maintenance.assert_called()
        want = [
            mock.call(
                'jenkins_slave_ps', 'Jenkins Slave Process', '/usr/lib/nagios/plugins/check_procs -c 1:1 -a slave.jar'
            )
        ]
        nrpe_instance_mock.add_check.assert_has_calls(want, any_order=True)

        nrpe_instance_mock.write.assert_called()

        want = [mock.call('nagios-nrpe.configured')]
        set_flag.assert_has_calls(want, any_order=True)

    def test_write_default_conf_update(self):
        jenkins_slave.write_default_conf('http://10.1.1.1:8080', self.user, self.group, self.conf_file)
        self.assertTrue(conf_match(self.conf_file, 'JENKINS_URL', 'http://10.1.1.1:8080'))

    def test_write_default_conf_reset(self):
        jenkins_slave.write_default_conf(None, self.user, self.group, self.conf_file)
        self.assertTrue(conf_match(self.conf_file, '#JENKINS_URL', ''))

    def test_file_to_units_executable_sh(self):
        source = os.path.join(self.charm_dir, 'files/download-slave.sh')
        dest = os.path.join(self.tmpdir, os.path.basename(source))
        jenkins_slave.file_to_units(source, dest, owner=self.user, group=self.group)
        with open(dest, 'rb') as fh:
            want = fh.read().decode('utf-8')
        self.assertTrue(conf_equals(source, want))
        self.assertEqual(pwd.getpwuid(os.stat(dest).st_uid).pw_name, self.user)
        self.assertEqual(grp.getgrgid(os.stat(dest).st_gid).gr_name, self.group)
        self.assertTrue(os.access(dest, os.X_OK))

    def test_file_to_units_executable_py(self):
        source = os.path.join(self.charm_dir, 'tests/unit/files/somefile.py')
        dest = os.path.join(self.tmpdir, os.path.basename(source))
        jenkins_slave.file_to_units(source, dest, owner=self.user, group=self.group)
        with open(dest, 'rb') as fh:
            want = fh.read().decode('utf-8')
        self.assertTrue(conf_equals(source, want))
        self.assertEqual(pwd.getpwuid(os.stat(dest).st_uid).pw_name, self.user)
        self.assertEqual(grp.getgrgid(os.stat(dest).st_gid).gr_name, self.group)
        self.assertTrue(os.access(dest, os.X_OK))

    def test_file_to_units_non_executable(self):
        source = os.path.join(self.charm_dir, 'files/jenkins-slave-logrotate-config')
        dest = os.path.join(self.tmpdir, os.path.basename(source))
        jenkins_slave.file_to_units(source, dest, owner=self.user, group=self.group)
        with open(dest, 'rb') as fh:
            want = fh.read().decode('utf-8')
        self.assertTrue(conf_equals(dest, want))
        self.assertEqual(pwd.getpwuid(os.stat(dest).st_uid).pw_name, self.user)
        self.assertEqual(grp.getgrgid(os.stat(dest).st_gid).gr_name, self.group)
        self.assertFalse(os.access(dest, os.X_OK))

    def test_file_to_units_non_executable_x_on_disk(self):
        source = os.path.join(self.charm_dir, 'tests/unit/files/somefile')
        dest = os.path.join(self.tmpdir, os.path.basename(source))
        jenkins_slave.file_to_units(source, dest, owner=self.user, group=self.group)
        with open(dest, 'rb') as fh:
            want = fh.read().decode('utf-8')
        self.assertTrue(conf_equals(source, want))
        self.assertEqual(pwd.getpwuid(os.stat(dest).st_uid).pw_name, self.user)
        self.assertEqual(grp.getgrgid(os.stat(dest).st_gid).gr_name, self.group)
        self.assertFalse(os.access(dest, os.X_OK))

    def test_file_to_units_with_perms(self):
        source = os.path.join(self.charm_dir, 'tests/unit/files/somefile')
        dest = os.path.join(self.tmpdir, os.path.basename(source))
        jenkins_slave.file_to_units(source, dest, owner=self.user, group=self.group, perms=0o640)
        with open(dest, 'rb') as fh:
            want = fh.read().decode('utf-8')
        self.assertTrue(conf_equals(source, want))
        self.assertEqual(pwd.getpwuid(os.stat(dest).st_uid).pw_name, self.user)
        self.assertEqual(grp.getgrgid(os.stat(dest).st_gid).gr_name, self.group)
        self.assertEqual(os.stat(dest)[stat.ST_MODE], 0o100640)


def conf_equals(conf_file, want):
    with open(conf_file, 'rb') as conf:
        got = conf.read().decode('utf-8')
        if got == want:
            return True
        print('{}\n != \n{}'.format(got, want))
    return False


def conf_match(conf_file, key, value):
    with open(conf_file, 'rb') as conf:
        for line in conf.readlines():
            line = line.decode('utf-8').rstrip('\n')
            if line == '{}="{}"'.format(key, value):
                return True
    return False


if __name__ == '__main__':
    unittest.main()
